/*
 *  Copyright 2010-2011, Plutext Pty Ltd.
 *   
 *  This file is part of docx4j.

    docx4j is licensed under the Apache License, Version 2.0 (the "License"); 
    you may not use this file except in compliance with the License. 

    You may obtain a copy of the License at 

        http://www.apache.org/licenses/LICENSE-2.0 

    Unless required by applicable law or agreed to in writing, software 
    distributed under the License is distributed on an "AS IS" BASIS, 
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
    See the License for the specific language governing permissions and 
    limitations under the License.

 */
package org.docx4j;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.JAXBElement;

import org.apache.log4j.Logger;
import org.docx4j.openpackaging.packages.WordprocessingMLPackage;
import org.docx4j.openpackaging.parts.WordprocessingML.CommentsPart;
import org.docx4j.openpackaging.parts.WordprocessingML.EndnotesPart;
import org.docx4j.openpackaging.parts.WordprocessingML.FooterPart;
import org.docx4j.openpackaging.parts.WordprocessingML.FootnotesPart;
import org.docx4j.openpackaging.parts.WordprocessingML.HeaderPart;
import org.docx4j.openpackaging.parts.WordprocessingML.MainDocumentPart;
import org.docx4j.openpackaging.parts.relationships.Namespaces;
import org.docx4j.openpackaging.parts.relationships.RelationshipsPart;
import org.docx4j.relationships.Relationship;
import org.docx4j.utils.CompoundTraversalUtilVisitorCallback;
import org.docx4j.utils.SingleTraversalUtilVisitorCallback;
import org.docx4j.utils.TraversalUtilVisitor;
import org.docx4j.wml.Body;
import org.docx4j.wml.CTFtnEdn;
import org.docx4j.wml.CTObject;
import org.docx4j.wml.Pict;
import org.docx4j.wml.Comments.Comment;

/**
 * Traverse a list of JAXB objects (eg document.xml's document/body children), and do something to them.
 * 
 * This is similar to what one could do via XSLT, but avoids marshalling/unmarshalling. The downside is that not
 * everything will necessarily get traversed here since visitChildren is not (yet) comprehensive.
 * 
 * @author jharrop, alberto
 * 
 */
public class TraversalUtil {

  private static Logger log = Logger.getLogger(TraversalUtil.class);

  public interface Callback {

    void walkJAXBElements(Object parent);

    List<Object> getChildren(Object o);

    /**
     * Visits a node in pre order (before its children have been visited).
     * 
     * A node is visited only if all its parents have been traversed ( {@link #shouldTraverse(Object)}).</p>
     * 
     * <p>
     * Implementations can have side effects.
     * </p>
     */
    List<Object> apply(Object o);

    /**
     * Decide whether this node's children should be traversed.
     * 
     * @return whether the children of this node should be visited
     */
    boolean shouldTraverse(Object o);

  }

  public static abstract class CallbackImpl implements Callback {

    // Depth first
    public void walkJAXBElements(final Object parent) {

      final List children = getChildren(parent);
      if (children != null) {

        for (Object o : children) {

          // if its wrapped in javax.xml.bind.JAXBElement, get its
          // value; this is ok, provided the results of the Callback
          // won't be marshalled
          o = XmlUtils.unwrap(o);

          this.apply(o);

          if (this.shouldTraverse(o)) {
            walkJAXBElements(o);
          }

        }
      }
    }

    public List<Object> getChildren(final Object o) {
      return TraversalUtil.getChildrenImpl(o);
    }

    /**
     * Visits a node in pre order (before its children have been visited).
     * 
     * A node is visited only if all its parents have been traversed ( {@link #shouldTraverse(Object)}).</p>
     * 
     * <p>
     * Implementations can have side effects.
     * </p>
     */
    public abstract List<Object> apply(Object o);

    /**
     * Decide whether this node's children should be traversed.
     * 
     * @return whether the children of this node should be visited
     */
    public boolean shouldTraverse(final Object o) {
      return true;
    }

  }

  Callback cb;

  public TraversalUtil(final Object parent, final Callback cb) {

    this.cb = cb;
    cb.walkJAXBElements(parent);
  }

  static void visitChildrenImpl(final Object o) {

  }

  public static List<Object> getChildrenImpl(final Object o) {

    log.debug("getting children of " + o.getClass().getName());
    if (o instanceof org.docx4j.wml.Text) {
      return null;
    }

    // Short circuit for common elements
    if (o instanceof List) {
      // Handy if you have your own list of objects you wish to process
      return (List<Object>) o;
    } else if (o instanceof org.docx4j.wml.ContentAccessor) {
      return ((org.docx4j.wml.ContentAccessor) o).getContent();
    } else if (o instanceof org.docx4j.wml.SdtElement) {
      return ((org.docx4j.wml.SdtElement) o).getSdtContent().getContent();
    } else if (o instanceof org.docx4j.dml.wordprocessingDrawing.Inline) {
      final org.docx4j.dml.wordprocessingDrawing.Inline inline = (org.docx4j.dml.wordprocessingDrawing.Inline) o;
      if (inline.getGraphic() != null) {
        log.info("found a:graphic");
        final org.docx4j.dml.Graphic graphic = inline.getGraphic();
        if (graphic.getGraphicData() != null) {
          final org.docx4j.dml.GraphicData graphicData = graphic.getGraphicData();
          // Its not graphicData.getAny() we're typically interested in
          if (graphicData.getPic() != null && graphicData.getPic().getBlipFill() != null
              && graphicData.getPic().getBlipFill().getBlip() != null) {
            log.info("found CTBlip");
            final List<Object> artificialList = new ArrayList<Object>();
            artificialList.add(graphicData.getPic().getBlipFill().getBlip());
            return artificialList;
          } else {
            // Chart is in here
            return graphicData.getAny();
          }

        }
      }
    } else if (o instanceof Pict) {
      return ((Pict) o).getAnyAndAny(); // (why didn't the reflection below find this?)
    } else if (o instanceof org.docx4j.vml.CTShape) {
      // return ((org.docx4j.vml.CTShape)o).getAny();
      final List<Object> artificialList = new ArrayList<Object>();
      for (final JAXBElement<?> j : ((org.docx4j.vml.CTShape) o).getPathOrFormulasOrHandles()) {
        artificialList.add(j);
      }
      return artificialList;
    } else if (o instanceof org.docx4j.vml.CTTextbox) {
      // return ((org.docx4j.vml.CTTextbox)o).getAny();
      return ((org.docx4j.vml.CTTextbox) o).getTxbxContent().getEGBlockLevelElts();
      // grandchildren

      // } else if (o instanceof org.docx4j.wml.CTTxbxContent) {
      // return ((org.docx4j.wml.CTTxbxContent)o).getEGBlockLevelElts();
    } else if (o instanceof CTObject) {
      return ((CTObject) o).getAnyAndAny();
    } else if (o instanceof HeaderPart) {
      return ((HeaderPart) o).getJaxbElement().getEGBlockLevelElts();
    } else if (o instanceof FooterPart) {
      return ((FooterPart) o).getJaxbElement().getEGBlockLevelElts();
    }

    // OK, what is this? Use reflection ..
    // This should work for things including w:drawing
    log.debug(".. looking for method which returns list ");
    try {
      final Method[] methods = o.getClass().getDeclaredMethods();
      for (int i = 0; i < methods.length; i++) {
        final Method m = methods[i];
        // ignore methods which need parameters
        if (m.getReturnType().getName().equals("java.util.List") && m.getParameterTypes().length == 0) {
          return (List<Object>) m.invoke(o);
        }
      }

    } catch (final Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    log.debug(".. no list member");
    return null;
  }

  public static void main(final String[] args) throws Exception {

    final String inputfilepath = System.getProperty("user.dir") + "/sample-docs/sample-docx.xml";

    final WordprocessingMLPackage wordMLPackage = WordprocessingMLPackage.load(new java.io.File(inputfilepath));
    final MainDocumentPart documentPart = wordMLPackage.getMainDocumentPart();

    final org.docx4j.wml.Document wmlDocumentEl = documentPart.getJaxbElement();
    final Body body = wmlDocumentEl.getBody();

    new TraversalUtil(body,

    new Callback() {

      String indent = "";

      @Override
      public List<Object> apply(final Object o) {

        String text = "";
        if (o instanceof org.docx4j.wml.Text) {
          text = ((org.docx4j.wml.Text) o).getValue();
        }

        System.out.println(indent + o.getClass().getName() + "  \"" + text + "\"");
        return null;
      }

      @Override
      public boolean shouldTraverse(final Object o) {
        return true;
      }

      // Depth first
      @Override
      public void walkJAXBElements(final Object parent) {

        indent += "    ";

        final List children = getChildren(parent);
        if (children != null) {

          for (Object o : children) {

            // if its wrapped in javax.xml.bind.JAXBElement, get its
            // value; this is ok, provided the results of the Callback
            // won't be marshalled
            o = XmlUtils.unwrap(o);

            this.apply(o);

            if (this.shouldTraverse(o)) {
              walkJAXBElements(o);
            }

          }
        }

        indent = indent.substring(0, indent.length() - 4);
      }

      @Override
      public List<Object> getChildren(final Object o) {
        return TraversalUtil.getChildrenImpl(o);
      }

    }

    );

  }

  public static void replaceChildren(final Object o, final List<Object> newChildren) {

    // Available if you need something like this. Would be used with
    // something like:

    /*
     * public void walkJAXBElements(Object parent){ // Breadth first
     * 
     * List<Object> newChildren = new ArrayList<Object>();
     * 
     * List children = getChildren(parent); if (children==null) { log.warn("no children: " + parent.getClass().getName()
     * ); } else { for (Object o : getChildren(parent) ) {
     * 
     * // if its wrapped in javax.xml.bind.JAXBElement, get its value o = XmlUtils.unwrap(o);
     * 
     * newChildren.addAll(this.apply(o)); } } // Replace list, so we'll traverse all the new sdts we've just created
     * replaceChildren(parent, newChildren);
     * 
     * 
     * for (Object o : getChildren(parent) ) {
     * 
     * this.apply(o);
     * 
     * if ( this.shouldTraverse(o) ) { walkJAXBElements(o); }
     * 
     * }
     * 
     * }
     */

    log.debug("Clearing " + o.getClass().getName());

    if (o instanceof org.docx4j.wml.ContentAccessor) {

      ((org.docx4j.wml.ContentAccessor) o).getContent().clear();
      ((org.docx4j.wml.ContentAccessor) o).getContent().addAll(newChildren);

    } else if (o instanceof org.docx4j.wml.SdtElement) {

      ((org.docx4j.wml.SdtElement) o).getSdtContent().getContent().clear();
      ((org.docx4j.wml.SdtElement) o).getSdtContent().getContent().addAll(newChildren);

    } else if (o instanceof org.docx4j.wml.CTTxbxContent) {

      ((org.docx4j.wml.CTTxbxContent) o).getEGBlockLevelElts().clear();
      ((org.docx4j.wml.CTTxbxContent) o).getEGBlockLevelElts().addAll(newChildren);

    } else {

      log.warn("Don't know how to replaceChildren in " + o.getClass().getName());

      if (o instanceof org.w3c.dom.Node) {
        log.warn(" IGNORED " + ((org.w3c.dom.Node) o).getNodeName());
      } else {
        // log.warn(" IGNORED " + o.getClass().getName());
      }

    }
  }

  public static void visit(final WordprocessingMLPackage wmlPackage, final boolean bodyOnly,
      final TraversalUtilVisitor visitor) {
    if (visitor != null) {
      visit(wmlPackage, bodyOnly, new SingleTraversalUtilVisitorCallback(visitor));
    }
  }

  public static void visit(final WordprocessingMLPackage wmlPackage, final boolean bodyOnly,
      final List<TraversalUtilVisitor> visitorList) {
    final CompoundTraversalUtilVisitorCallback callback = null;
    if (visitorList != null && !visitorList.isEmpty()) {
      if (visitorList.size() > 1) {
        visit(wmlPackage, bodyOnly, new CompoundTraversalUtilVisitorCallback(visitorList));
      } else {
        visit(wmlPackage, bodyOnly, visitorList.get(0));
      }
    }
  }

  public static void visit(final WordprocessingMLPackage wmlPackage, final boolean bodyOnly, final Callback callback) {
    MainDocumentPart mainDocument = null;
    RelationshipsPart relPart = null;
    List<Relationship> relList = null;
    List<Object> elementList = null;
    if (wmlPackage != null && callback != null) {
      mainDocument = wmlPackage.getMainDocumentPart();
      callback.walkJAXBElements(mainDocument.getJaxbElement().getBody());
      if (!bodyOnly) {
        relPart = mainDocument.getRelationshipsPart();
        relList = relPart.getRelationships().getRelationship();
        for (final Relationship rs : relList) {
          elementList = null;
          if (Namespaces.HEADER.equals(rs.getType())) {
            elementList = ((HeaderPart) relPart.getPart(rs)).getJaxbElement().getEGBlockLevelElts();
          } else if (Namespaces.FOOTER.equals(rs.getType())) {
            elementList = ((FooterPart) relPart.getPart(rs)).getJaxbElement().getEGBlockLevelElts();
          } else if (Namespaces.ENDNOTES.equals(rs.getType())) {
            elementList = new ArrayList();
            for (final CTFtnEdn endnote : ((EndnotesPart) relPart.getPart(rs)).getJaxbElement().getEndnote()) {
              elementList.addAll(endnote.getEGBlockLevelElts());
            }
          } else if (Namespaces.FOOTNOTES.equals(rs.getType())) {
            elementList = new ArrayList();
            for (final CTFtnEdn footnote : ((FootnotesPart) relPart.getPart(rs)).getJaxbElement().getFootnote()) {
              elementList.addAll(footnote.getEGBlockLevelElts());
            }
          } else if (Namespaces.COMMENTS.equals(rs.getType())) {
            elementList = new ArrayList();
            for (final Comment comment : ((CommentsPart) relPart.getPart(rs)).getJaxbElement().getComment()) {
              elementList.addAll(comment.getEGBlockLevelElts());
            }
          }
          if (elementList != null && !elementList.isEmpty()) {
            System.out.println("Processing target: " + rs.getTarget() + ", type: " + rs.getType());
            callback.walkJAXBElements(elementList);
          }
        }
      }
    }
  }

  public static void visit(final Object parent, final TraversalUtilVisitor visitor) {
    if (visitor != null) {
      visit(parent, new SingleTraversalUtilVisitorCallback(visitor));
    }
  }

  public static void visit(final Object parent, final List<TraversalUtilVisitor> visitorList) {
    final CompoundTraversalUtilVisitorCallback callback = null;
    if (visitorList != null && !visitorList.isEmpty()) {
      if (visitorList.size() > 1) {
        visit(parent, new CompoundTraversalUtilVisitorCallback(visitorList));
      } else {
        visit(parent, visitorList.get(0));
      }
    }
  }

  public static void visit(final Object parent, final Callback callback) {
    if (parent != null && callback != null) {
      callback.walkJAXBElements(parent);
    }
  }

  // private void describeDrawing( org.docx4j.wml.Drawing d) {
  //			
  // log.info("In wml.Drawing" );
  //			
  // if ( d.getAnchorOrInline().get(0) instanceof Anchor ) {
  //				
  // log.info(" ENCOUNTERED w:drawing/wp:anchor " );
  // // That's all for now...
  //				
  // } else if ( d.getAnchorOrInline().get(0) instanceof Inline ) {
  //				
  // // Extract
  // w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip/@r:embed
  //				
  // Inline inline = (Inline )d.getAnchorOrInline().get(0);
  //				
  // Pic pic = inline.getGraphic().getGraphicData().getPic();
  //					
  // if (pic!=null) {
  // log.info("image relationship: " + pic.getBlipFill().getBlip().getEmbed()
  // );
  // }
  //				
  //				
  // } else {
  //				
  // log.info(" Didn't get Inline :(  How to handle " +
  // d.getAnchorOrInline().get(0).getClass().getName() );
  // }
  //			
  // }

}
